import os

from shiny import run_app

from paos import logger

from paos import __pkg_name__
from paos import __version__


def main():
    """
    This function is the entry point for the command line interface of the PAOS GUI.
    It sets up the logging and the app path, and then runs the shiny app.

    Parameters
    ----------
    None

    Returns
    -------
    None
    """
    app = os.path.realpath(os.path.dirname(__file__)) + "/app.py"
    logger.info(f"Running app: {__pkg_name__} GUI v{__version__}")
    run_app(app, reload=True, launch_browser=True)


if __name__ == "__main__":
    main()



import os
import time
import configparser
from pathlib import Path
import matplotlib.pyplot as plt

from htmltools import Tag
from starlette.requests import Request as StarletteRequest

from shiny import App
from shiny import render
from shiny import ui
from shiny import reactive
from shiny import req
from shiny.types import FileInfo

import paos
from paos import __pkg_name__
from paos import __version__
from paos import __author__
from paos import __license__
from paos.core.parseConfig import parse_config

from paos.gui.core.io import to_ini
from paos.gui.core.elems import app_elems
from paos.gui.core.plot import simple_plot
from paos.gui.core.plot import zernike_plot
from paos.gui.core.shared import ICONS
from paos.gui.core.shared import nested_div
from paos.gui.core.shared import menu_panel
from paos.gui.core.shared import refresh_ui
from paos.gui.core.shared import modal_download


def app_ui(request: StarletteRequest) -> Tag:
    return ui.page_navbar(
        ui.nav_spacer(),
        ui.nav_panel(
            "System Explorer",
            ui.layout_columns(
                *[
                    ui.card(ui.card_header("General"), nested_div("general")),
                    ui.card(ui.card_header("Units"), nested_div("units")),
                    ui.card(ui.card_header("Simulation"), nested_div("sim")),
                    ui.card(
                        ui.card_header(
                            ui.layout_columns(
                                {"style": "text-align: center;"},
                                ui.p("#"),
                                ui.p("Field"),
                            ),
                        ),
                        nested_div("field"),
                    ),
                    ui.card(
                        ui.card_header(
                            ui.layout_columns(
                                {"style": "text-align: center;"},
                                ui.p("#"),
                                ui.p("Wavelength"),
                            ),
                        ),
                        nested_div("wl"),
                    ),
                ],
                open=False,
            ),
        ),
        ui.nav_panel(
            "Lens Editor",
            ui.card(
                nested_div("lens"),
            ),
        ),
        ui.nav_panel(
            "Zernike Editor",
            ui.layout_columns(
                ui.div(
                    ui.card(nested_div("zernike_explorer"), full_screen=True),
                ),
                ui.card(nested_div("zernike_plots"), full_screen=True),
            ),
        ),
        ui.nav_panel(
            "Analysis",
            ui.layout_columns(
                ui.div(
                    ui.card(nested_div("pop"), full_screen=True),
                    ui.card(nested_div("raytrace"), full_screen=True),
                ),
                ui.card(nested_div("plots"), full_screen=True),
            ),
        ),
        id="navbar",
        title=ui.tags.div(
            ui.tags.a(
                ui.tags.img(src="static/logo.png", height="50px"),
                href="https://github.com/arielmission-space/PAOS",
            ),
            id="logo-top",
            class_="navigation-logo",
        ),
        header=ui.page_navbar(
            [
                ui.nav_menu(
                    "File",
                    menu_panel("open"),
                    menu_panel("save"),
                    menu_panel("close"),
                ),
                # ui.nav_menu(
                #     "Edit",
                #     menu_panel("load"),
                #     menu_panel("refresh"),
                # ),
                ui.nav_menu(
                    "Help",
                    menu_panel("docs"),
                    menu_panel("about"),
                ),
            ],
        ),
        footer=ui.panel_well(
            ui.p(
                f"{__pkg_name__} v{__version__}; {__author__}",
            ),
        ),
        window_title=f"{__pkg_name__} GUI",
        # selected="System",
    )


def server(input, output, session):
    ini_file = reactive.value("filename")
    config = reactive.value(configparser.ConfigParser())
    retval = reactive.value({})
    figure = reactive.value(None)
    figure_zernike = reactive.value(None)

    for file in os.listdir(cache):
        os.remove(os.path.join(cache, file))

    @reactive.effect
    @reactive.event(input.save, input.calc_raytrace, input.calc_pop)
    def _():
        req(input.save)
        req(input.calc_raytrace)
        to_ini(input=input, config=config, tmp=cache / "tmp.ini")

    @reactive.calc
    def calc_raytrace():
        req(config.get().sections())
        req(input.raytrace_select_field())
        req(input.raytrace_select_wl())

        field = input.raytrace_select_field()
        wl = input.raytrace_select_wl()

        field_idx = int(field[1:]) - 1
        wl_idx = int(wl[1:]) - 1

        _, _, _, fields, opt_chains = parse_config(cache / "tmp.ini")

        with ui.Progress(min=1, max=15) as p:
            p.set(message="Calculation in progress", detail="")
            raytrace = paos.raytrace(fields[field_idx], opt_chains[wl_idx])

            p.set(15, message="Raytrace complete!", detail="")
            time.sleep(1.0)

        return "\n".join(raytrace)

    @reactive.effect
    @reactive.event(input.download_raytrace)
    def download_raytrace():
        req(input.calc_raytrace())
        req(input.download_raytrace())
        req(config.get().sections())
        modal_download("raytrace", "txt")

    @render.download
    def download_raytrace_txt():
        raytrace: str = calc_raytrace()
        outfile: list[FileInfo] | None = input.save_txt()
        print(f"Downloaded {outfile}")

        with open(cache / outfile, "w") as f:
            f.write(raytrace)

        return os.path.join(os.path.dirname(__file__), "cache", outfile)

    @render.text
    @reactive.event(input.calc_raytrace)
    def raytrace_inputs():
        req(input.calc_raytrace())

        refresh_ui(
            "raytrace_inputs",
            [ui.output_text_verbatim("raytrace_inputs", placeholder=True)],
        )

        field = input.raytrace_select_field()
        wl = input.raytrace_select_wl()

        return f"Field: {field}, Wavelength: {wl}"

    @render.text
    @reactive.event(input.calc_raytrace)
    def raytrace_output():
        req(input.calc_raytrace())

        refresh_ui(
            "raytrace_output",
            [ui.output_text_verbatim("raytrace_output", placeholder=True)],
        )

        return calc_raytrace()

    @reactive.calc
    def calc_pop():
        req(config.get().sections())
        req(input.pop_select_field())
        req(input.pop_select_wl())

        field = input.pop_select_field()
        wl = input.pop_select_wl()

        field_idx = int(field[1:]) - 1
        wl_idx = int(wl[1:]) - 1

        pup_diameter, parameters, wavelengths, fields, opt_chains = parse_config(
            cache / "tmp.ini"
        )

        with ui.Progress(min=1, max=15) as p:
            p.set(message="Calculation in progress", detail="")
            retval.set(
                paos.run(
                    pup_diameter,
                    1.0e-6 * wavelengths[wl_idx],
                    parameters["grid_size"],
                    parameters["zoom"],
                    fields[field_idx],
                    opt_chains[wl_idx],
                )
            )

            p.set(15, message="POP complete!", detail="")
            time.sleep(1.0)

        return "POP output available: try plotting it!"

    @reactive.effect
    @reactive.event(input.download_pop)
    def download_pop():
        req(input.calc_pop())
        req(input.download_pop())
        req(config.get().sections())
        modal_download("pop", "h5")

    @render.download
    def download_pop_h5():
        outfile: list[FileInfo] | None = input.save_h5()

        paos.save_output(
            retval.get(),
            str(cache / outfile),
            keys_to_keep=["amplitude", "dx", "dy", "wl"],
        )

        return os.path.join(os.path.dirname(__file__), "cache", outfile)

    @render.text
    @reactive.event(input.calc_pop)
    def pop_inputs():
        req(input.calc_pop())

        refresh_ui(
            "pop_inputs",
            [ui.output_text_verbatim("pop_inputs", placeholder=True)],
        )

        field = input.pop_select_field()
        wl = input.pop_select_wl()

        return f"Field: {field}, Wavelength: {wl}"

    @render.text
    @reactive.event(input.calc_pop)
    def pop_output():
        req(input.calc_pop())

        refresh_ui(
            "pop_output",
            [ui.output_text_verbatim("pop_output", placeholder=True)],
        )

        return calc_pop()

    @render.plot(alt="PSF plot")
    @reactive.event(input.do_plot)
    def plot():
        req(input.do_plot())
        req(input.calc_pop())
        req(input.plot_select_surface())
        req(input.plot_select_scale())
        req(input.plot_select_zoom())

        surface = int(input.plot_select_surface()[1:])
        scale = input.plot_select_scale()
        zoom = input.plot_select_zoom()
        dark_rings = input.plot_select_dark_rings()

        fig, ax = plt.subplots()
        simple_plot(
            fig=fig,
            axis=ax,
            key=surface,
            item=retval.get()[surface],
            ima_scale=scale,
            options={
                surface: {
                    "surface_zoom": float(zoom),
                    "dark_rings": bool(dark_rings),
                }
            },
        )

        figure.set(fig)

        return fig

    @reactive.effect
    @reactive.event(input.download_plot)
    def download_plot():
        req(input.do_plot())
        req(input.download_plot())
        req(config.get().sections())
        req(retval.get())
        modal_download("plot", "pdf")

    @render.download
    def download_plot_pdf():
        outfile: list[FileInfo] | None = input.save_pdf()

        figure.get().savefig(cache / outfile)

        return os.path.join(os.path.dirname(__file__), "cache", outfile)

    @render.text
    @reactive.event(input.do_plot)
    def plot_inputs():
        req(input.do_plot())

        refresh_ui(
            "plot_inputs",
            [ui.output_text_verbatim("plot_inputs", placeholder=True)],
        )

        surface = input.plot_select_surface()
        scale = input.plot_select_scale()
        zoom = input.plot_select_zoom()
        dark_rings = input.plot_select_dark_rings()

        return f"Surface: {surface}, Scale: {scale}, Zoom: {zoom}, Dark Rings: {dark_rings}"

    @reactive.effect
    @reactive.event(input.save)
    def save():
        req(config.get().sections())
        modal_download("config", "ini")

    @render.download
    def download_config_ini():
        outfile: list[FileInfo] | None = input.save_ini()

        with open(cache / "tmp.ini", "r") as f:
            with open(cache / outfile, "w") as cf:
                cf.write(f.read())

        return os.path.join(os.path.dirname(__file__), "cache", outfile)

    @reactive.effect
    @reactive.event(input.docs)
    def _():
        req(input.docs())
        m = ui.modal(
            ui.markdown(
                f"Click [here](https://paos.readthedocs.io/en/latest/) to access the {__pkg_name__} documentation."
            ),
            title="Documentation",
            easy_close=True,
        )
        ui.modal_show(m)

    @reactive.effect
    @reactive.event(input.about)
    def about():
        req(input.about())
        m = ui.modal(
            ui.markdown(
                f"SOFTWARE: {__pkg_name__} v{__version__}  \n"
                f"AUTHOR: {__author__}  \n"
                f"LICENSE: {__license__}  \n"
            ),
            title="About",
            easy_close=True,
        )
        ui.modal_show(m)

    @reactive.effect
    @reactive.event(input.docs)
    def docs():
        req(input.docs())
        m = ui.modal(
            ui.markdown(
                """
                Click [here](https://paos.readthedocs.io/en/latest/) to access the PAOS documentation.
                """
            ),
            title="Documentation",
            easy_close=True,
        )
        ui.modal_show(m)

    @reactive.effect
    @reactive.event(input.open)
    def _():
        req(input.open())
        m = ui.modal(
            ui.input_file(
                id="open_ini",
                label=ui.markdown(
                    "Input files must be in the INI format.  \n"
                    "Example files can be found in the PAOS public [GitHub repository](https://github.com/arielmission-space/PAOS)."
                ),
                accept=[".ini"],
                multiple=False,
                button_label="Browse",
            ),
            title="Open INI File",
            easy_close=True,
            footer=ui.markdown(
                "Note: you may only open one file per session.  \n"
                "Refresh the page to open a different file."
            ),
        )
        ui.modal_show(m)

    @render.text
    @reactive.event(input.open_ini, input.zernike_select_surface)
    def zernike_inputs():
        req(config.get().sections())
        req(input.zernike_select_surface())

        refresh_ui(
            "zernike_inputs",
            [ui.output_text_verbatim("zernike_inputs", placeholder=True)],
        )

        surface = input.zernike_select_surface()

        (_, _, _, _, _, _, _, zernike_elems, _, _, _, _) = app_elems(config.get())

        surface_key = int(surface[1:])
        refresh_ui("zernike", zernike_elems, mode="nested-dict", key=surface_key)

        return f"Surface: {surface}"

    @render.plot(alt="Zernike plot")
    @reactive.event(input.do_plot_zernike)
    def plot_zernike():
        req(input.do_plot_zernike())
        req(input.zernike_plot_select_surface())
        req(config.get().sections())

        to_ini(input=input, config=config, tmp=cache / "tmp.ini")

        surface = input.zernike_plot_select_surface()
        surface_key = int(surface[1:])
        zernike_section = f"lens_{surface_key:02d}"
        zernike_section = config.get()[zernike_section]

        zindex = zernike_section.get("zindex").split(",")
        zindex = list(map(int, zindex))
        zcoeffs = zernike_section.get("z").split(",")
        zcoeffs = list(map(float, zcoeffs))
        wavelength = float(zernike_section.get("par1"))
        ordering = zernike_section.get("par2")
        normalize = bool(zernike_section.get("par3"))

        fig, ax = plt.subplots()
        zernike_plot(
            fig=fig,
            axis=ax,
            surface=surface,
            index=zindex,
            Z=zcoeffs,
            wavelength=wavelength,
            ordering=ordering,
            normalize=normalize,
            grid_size=int(input.grid_size()),
        )

        figure_zernike.set(fig)

        return fig

    @reactive.effect
    @reactive.event(input.download_plot_zernike)
    def download_plot_zernike():
        req(config.get().sections())
        req(input.do_plot_zernike())
        modal_download("plot_zernike", "pdf")

    @render.download
    def download_plot_zernike_pdf():
        outfile: list[FileInfo] | None = input.save_pdf()

        figure_zernike.get().savefig(cache / outfile)

        return os.path.join(os.path.dirname(__file__), "cache", outfile)

    @render.text
    @reactive.event(input.do_plot_zernike)
    def plot_zernike_inputs():
        req(input.do_plot_zernike())

        refresh_ui(
            "plot_zernike_inputs",
            [ui.output_text_verbatim("plot_zernike_inputs", placeholder=True)],
        )

        surface = input.zernike_plot_select_surface()

        return f"Surface: {surface}"

    @reactive.effect
    @reactive.event(input.open_ini)
    # async def open_ini():
    def open_ini():
        req(input.open_ini())
        file: list[FileInfo] | None = input.open_ini()

        if file is None:
            return

        if not file[0]["name"].endswith(".ini"):
            print("Invalid file")
            return

        if ini_file.get().endswith(".ini") and ini_file.get() != file[0]["name"]:
            return
            # await session.send_custom_message("refresh", "")

        ini_file.set(file[0]["datapath"])
        config.get().read(ini_file.get())

        (
            general_elems,
            units_elems,
            sim_elems,
            field_elems,
            wl_elems,
            lens_elems,
            zernike_explorer_elems,
            zernike_elems,
            zernike_plots_elems,
            pop_elems,
            raytrace_elems,
            plots_elems,
        ) = app_elems(config.get())

        refresh_ui("general", general_elems)
        refresh_ui("units", units_elems)
        refresh_ui("sim", sim_elems)
        refresh_ui("field", field_elems, mode="body")
        refresh_ui("wl", wl_elems, mode="body")
        refresh_ui("lens", lens_elems, mode="dict")
        refresh_ui("zernike_explorer", zernike_explorer_elems)
        if zernike_elems:
            for key in zernike_elems.keys():
                refresh_ui("zernike", zernike_elems, mode="nested-dict", key=key)
        refresh_ui("zernike_plots", zernike_plots_elems)
        refresh_ui("pop", pop_elems)
        refresh_ui("raytrace", raytrace_elems)
        refresh_ui("plots", plots_elems)

    @reactive.effect
    @reactive.event(input.close)
    async def _():
        await session.close()


app_dir = Path(__file__).parent
www = app_dir / "www"
cache = app_dir / "cache"
Path(www).mkdir(exist_ok=True, parents=True)
Path(cache).mkdir(exist_ok=True, parents=True)
app = App(app_ui, server, debug=False, static_assets=www)



from shiny import ui

from paos import Zernike

from paos.gui.core.shared import ellipsis
from paos.gui.core.shared import output_text_verbatim
from paos.gui.core.shared import ICONS
from paos.gui.core.shared import card_header_class_
from paos.gui.core.shared import nested_div


def app_elems(config):
    general_elems = [
        ui.input_text(
            "project",
            "Project name",
            value=config["general"].get("project", "Template"),
        ),
        ui.input_text("comment", "Comment", value=config["general"].get("comment", "")),
        ui.input_text(
            "version",
            "Version",
            value=config["general"].get("version", "0.1"),
        ),
    ]

    units_elems = [
        ui.p("Lens: ", config["general"].get("lens_unit", "m")),
        ui.p("Angle: ", config["general"].get("angle_unit", "degrees")),
        ui.p(
            "Wavelength: ",
            config["general"].get("wavelength_units", "micrometers"),
        ),
        ui.p("Temperature: ", config["general"].get("temperature_units", "C")),
        ui.p("Pressure: ", config["general"].get("pressure_units", "atm")),
    ]

    sim_elems = [
        ui.input_text(
            "grid_size",
            "Grid size",
            value=config["general"].getint("grid_size", "512"),
        ),
        ui.input_text("zoom", "Zoom", value=config["general"].getint("zoom", "4")),
        ui.input_text(
            "tambient",
            "Ambient temperature",
            value=config["general"].getfloat("tambient", "20.0"),
        ),
        ui.input_text(
            "pambient",
            "Ambient pressure",
            value=config["general"].getfloat("pambient", "1.0"),
        ),
    ]

    field_elems = {}
    for n, (key, value) in enumerate(config.items("fields")):
        n += 1
        field_elems[n] = {}
        field_elems[n]["#"] = {
            "f": ui.p,
            "width": 5,
            "value": n,
        }
        field_elems[n]["Field"] = {
            "f": ui.input_text,
            "id": key,
            "value": value,
            "width": 7,
        }
    field_choices = [f"f{n}" for n in field_elems]

    wl_elems = {}
    for n, (key, value) in enumerate(config.items("wavelengths")):
        n += 1
        wl_elems[n] = {}
        wl_elems[n]["#"] = {
            "f": ui.p,
            "width": 5,
            "value": n,
        }
        wl_elems[n]["Wavelength"] = {
            "f": ui.input_text,
            "id": key,
            "value": value,
            "width": 7,
        }
    wl_choices = [f"w{n}" for n in wl_elems]

    lens_elems = {}
    for key in config.sections():
        if not key.startswith("lens_"):
            continue
        n = int(key.split("_")[1])
        item = config[key]

        lens_elems[n] = {}
        lens_elems[n]["#"] = {
            "f": ui.p,
            "width": 1,
            "value": n,
        }
        lens_elems[n]["SurfaceType"] = {
            "f": ui.input_select,
            "choices": [
                "INIT",
                "Coordinate Break",
                "Standard",
                "Paraxial Lens",
                "ABCD",
                "Zernike",
                "PSD",
                "Grid Sag",
            ],
            "width": 1,
            "selected": item.get("surfacetype"),
            "prefix": "lens_",
        }
        lens_elems[n]["Comment"] = {
            "f": ui.input_text,
            "value": item.get("comment"),
            "width": 1,
            "prefix": "lens_",
        }
        lens_elems[n]["Radius"] = {
            "f": ui.input_text,
            "value": item.get("radius", ""),
            "width": 1,
            "prefix": "lens_",
        }
        lens_elems[n]["Thickness"] = {
            "f": ui.input_text,
            "value": item.get("thickness", ""),
            "width": 1,
            "prefix": "lens_",
        }
        lens_elems[n]["Material"] = {
            "f": ui.input_select,
            "choices": [
                "",
                "MIRROR",
                "BK7",
                "BAF2",
                "CAF2",
                "SAPPHIRE",
                "SF11",
                "ZNSE",
            ],
            "width": 1,
            "selected": item.get("material", ""),
            "prefix": "lens_",
        }
        lens_elems[n]["Save"] = {
            "f": ui.input_checkbox,
            "value": item.getboolean("save", True),
            "width": 1,
            "prefix": "lens_",
        }
        lens_elems[n]["Ignore"] = {
            "f": ui.input_checkbox,
            "value": item.getboolean("ignore", False),
            "width": 1,
            "prefix": "lens_",
        }
        lens_elems[n]["Stop"] = {
            "f": ui.input_checkbox,
            "value": item.getboolean("stop", False),
            "width": 1,
            "prefix": "lens_",
        }
        aperture = item.get("aperture", "")
        if aperture != "":
            aperture = " ".join(aperture.split())
            aperture = aperture.split(",")
            aperture_type = str(aperture[0])
            xhw = float(aperture[1])
            yhw = float(aperture[2])
            xdecenter = float(aperture[3])
            ydecenter = float(aperture[4])
        else:
            aperture_type = xhw = yhw = xdecenter = ydecenter = ""
        lens_elems[n]["Aperture"] = {
            "title": "Settings",
            "f": ui.popover,
            "width": 1,
            "value": {
                "Aperture_Type": {
                    "f": ui.input_select,
                    "selected": aperture_type,
                    "choices": [
                        "",
                        "rectangular aperture",
                        "rectangular obscuration",
                        "elliptical aperture",
                        "elliptical obscuration",
                    ],
                    "label": "Type",
                    "prefix": "lens_",
                },
                "Aperture_xhw": {
                    "f": ui.input_text,
                    "value": xhw,
                    "label": "X-Half Width",
                    "prefix": "lens_",
                },
                "Aperture_yhw": {
                    "f": ui.input_text,
                    "value": yhw,
                    "label": "Y-Half Width",
                    "prefix": "lens_",
                },
                "Aperture_xdecenter": {
                    "f": ui.input_text,
                    "value": xdecenter,
                    "label": "X-Decenter",
                    "prefix": "lens_",
                },
                "Aperture_ydecenter": {
                    "f": ui.input_text,
                    "value": ydecenter,
                    "label": "Y-Decenter",
                    "prefix": "lens_",
                },
            },
        }
        lens_elems[n]["Par1"] = {
            "f": ui.input_text,
            "value": item.get("par1", ""),
            "width": 1,
            "prefix": "lens_",
        }
        lens_elems[n]["Par2"] = {
            "f": ui.input_text,
            "value": item.get("par2", ""),
            "width": 1,
            "prefix": "lens_",
        }
        lens_elems[n]["Par3"] = {
            "f": ui.input_text,
            "value": item.get("par3", ""),
            "width": 1,
            "prefix": "lens_",
        }
        lens_elems[n]["Par4"] = {
            "f": ui.input_text,
            "value": item.get("par4", ""),
            "width": 1,
            "prefix": "lens_",
        }
        lens_elems[n]["Par5"] = {
            "f": ui.input_text,
            "value": item.get("par5", ""),
            "width": 1,
            "prefix": "lens_",
        }
        lens_elems[n]["Par6"] = {
            "f": ui.input_text,
            "value": item.get("par6", ""),
            "width": 1,
            "prefix": "lens_",
        }
        lens_elems[n]["Par7"] = {
            "f": ui.input_text,
            "value": item.get("par7", ""),
            "width": 1,
            "prefix": "lens_",
        }
        lens_elems[n]["Par8"] = {
            "f": ui.input_text,
            "value": item.get("par8", ""),
            "width": 1,
            "prefix": "lens_",
        }

    zernike_elems = {}
    for key in config.sections():
        if not key.startswith("lens_"):
            continue
        item = config[key]
        if not item.get("surfacetype") == "Zernike":
            continue

        n = int(key.split("_")[1])
        zernike_elems[n] = {}

        zindex = item.get("zindex", "").split(",")
        zcoeff = item.get("z", "").split(",")

        ordering = item.get("par2")
        azimuthal, radial = Zernike.j2mn(N=len(zindex), ordering=ordering)

        for zi, zc in zip(zindex, zcoeff):
            zernike_elems[n][zi] = {}
            zernike_elems[n][zi]["n"] = {
                "f": ui.p,
                "width": 3,
                "value": radial[int(zi)],
            }
            zernike_elems[n][zi]["m"] = {
                "f": ui.p,
                "width": 3,
                "value": azimuthal[int(zi)],
            }
            zernike_elems[n][zi]["Zindex"] = {
                "f": ui.p,
                "width": 3,
                "value": zi,
            }
            zernike_elems[n][zi]["Zcoeff"] = {
                "f": ui.input_text,
                "width": 3,
                "value": zc,
                "prefix": f"lens_{n}_",
            }

    if zernike_elems:
        zernike_choices = [f"S{n}" for n in zernike_elems]
        zernike_explorer_elems = [
            ui.card_header(
                "Zernike Explorer",
                ui.popover(
                    ICONS["ellipsis"],
                    *[
                        ui.input_select(
                            id="zernike_select_surface",
                            label="Choose surface",
                            choices=zernike_choices,
                            selected=zernike_choices[0],
                        ),
                    ],
                    title="",
                    placement="top",
                ),
                class_=card_header_class_,
            ),
            output_text_verbatim("zernike_inputs"),
            nested_div("zernike"),
            ui.card_footer(),
        ]
        zernike_plots_elems = [
            ui.card_header(
                "Zernike Errormap",
                ui.popover(
                    ICONS["ellipsis"],
                    *[
                        ui.input_select(
                            id="zernike_plot_select_surface",
                            label="Choose surface",
                            choices=zernike_choices,
                            selected=zernike_choices[0],
                        ),
                    ],
                    title="",
                    placement="top",
                ),
                class_=card_header_class_,
            ),
            output_text_verbatim("plot_zernike_inputs"),
            ui.output_plot("plot_zernike"),
            ui.card_footer(
                ui.input_action_button("do_plot_zernike", "Run", icon=ICONS["run"]),
                ui.input_action_button(
                    "download_plot_zernike", "Download", icon=ICONS["save"]
                ),
            ),
        ]

    else:
        zernike_explorer_elems = []
        zernike_plots_elems = []

    pop_elems = [
        ui.card_header(
            "Fresnel POP",
            ellipsis(
                "pop",
                names=["field", "wl"],
                choices=[field_choices, wl_choices],
            ),
            class_=card_header_class_,
        ),
        output_text_verbatim("pop_inputs"),
        output_text_verbatim("pop_output"),
        ui.card_footer(
            ui.input_action_button("calc_pop", "Run", icon=ICONS["run"]),
            ui.input_action_button("download_pop", "Download", icon=ICONS["save"]),
        ),
    ]

    raytrace_elems = [
        ui.card_header(
            "Ray Tracing",
            ellipsis(
                "raytrace",
                names=["field", "wl"],
                choices=[field_choices, wl_choices],
            ),
            class_=card_header_class_,
        ),
        output_text_verbatim("raytrace_inputs"),
        output_text_verbatim("raytrace_output"),
        ui.card_footer(
            ui.input_action_button("calc_raytrace", "Run", icon=ICONS["run"]),
            ui.input_action_button("download_raytrace", "Download", icon=ICONS["save"]),
        ),
    ]

    surface_choices = [f"S{n}" for n in lens_elems if lens_elems[n]["Save"]["value"]]
    plots_elems = [
        ui.card_header(
            "Plots",
            ui.popover(
                ICONS["ellipsis"],
                *[
                    ui.input_select(
                        id="plot_select_surface",
                        label="Choose surface",
                        choices=surface_choices,
                        selected=surface_choices[-1],
                    ),
                    ui.input_select(
                        id="plot_select_scale",
                        label="Choose scale",
                        choices=["log", "linear"],
                    ),
                    ui.input_text(
                        id="plot_select_zoom",
                        label="Choose zoom",
                        value=1.0,
                    ),
                    ui.input_checkbox(
                        id="plot_select_dark_rings",
                        label="Dark rings",
                        value=True,
                    ),
                ],
                title="",
                placement="top",
            ),
            class_=card_header_class_,
        ),
        output_text_verbatim("plot_inputs"),
        ui.output_plot("plot"),
        ui.card_footer(
            ui.input_action_button("do_plot", "Run", icon=ICONS["run"]),
            ui.input_action_button("download_plot", "Download", icon=ICONS["save"]),
        ),
    ]

    return (
        general_elems,
        units_elems,
        sim_elems,
        field_elems,
        wl_elems,
        lens_elems,
        zernike_explorer_elems,
        zernike_elems,
        zernike_plots_elems,
        pop_elems,
        raytrace_elems,
        plots_elems,
    )



import numpy as np
import sympy
import astropy.units as u

from paos import logger


class PSD:
    """
    Generates a surface error field (SFE) with a given power spectral density (PSD) and surface roughness (SR).

    The PSD is given by the following expression:

    :math:`\\displaystyle PSD(f) = \\frac{A}{B + (f/f_{knee})^C}`

    Parameters
    ----------
    pupil : array like
        The pupil for which the SFE is generated.
    A : float
        The amplitude of the PSD.
    B : float
        PSD parameter. If B = 0, the PSD is a power law.
    C : float
        PSD parameter. It sets the slope of the PSD.
    f : array like
        The frequency grid.
    fknee : float
        The knee frequency of the PSD.
    fmin : float
        The minimum frequency of the PSD.
    fmax : float
        The maximum frequency of the PSD.
    SR : float
        The rms of the surface roughness.
    units : astropy.units
        The units of the SFE. Default is meters.

    Returns
    -------
    wfe : array like
        The generated surface error field.

    Example
    -------
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> import astropy.units as u

    >>> phi_x = 110.0  # mm
    >>> phi_y = 73.0  # mm
    >>> zoom = 4
    >>> wl = 0.550
    >>> D = zoom * np.max([phi_x, phi_y])
    >>> grid = 1024
    >>> delta = D / grid

    >>> A, B, C, fknee, fmin, fmax = 7.0, 0.0, 1.5, 1.0, 1 / 20, 1 / 2
    >>> SR = 0.0

    >>> x = y = np.arange(-grid // 2, grid // 2) * delta
    >>> xx, yy = np.meshgrid(x, y)
    >>> pupil = np.zeros((grid, grid))

    >>> mask = (2 * xx / phi_x) ** 2 + (2 * yy / phi_y) ** 2 <= 1
    >>> pupil[mask] = 1.0
    >>> wfo = np.ma.masked_array(pupil, mask=~mask)

    >>> fx = np.fft.fftfreq(wfo.shape[0], delta)
    >>> fxx, fyy = np.meshgrid(fx, fx)
    >>> f = np.sqrt(fxx**2 + fyy**2)
    >>> f[f == 0] = 1e-100

    >>> wfo_ = PSD(wfo, A, B, C, f, fknee, fmin, fmax, SR, units=u.nm)

    >>> plt.figure()
    >>> plt.imshow(wfo_(), cmap="jet", origin="lower")
    >>> plt.xlim(grid // 2 - grid // (2 * zoom), grid // 2 + grid // (2 * zoom))
    >>> plt.ylim(grid // 2 - grid // (2 * zoom), grid // 2 + grid // (2 * zoom))
    >>> plt.colorbar()
    >>> plt.show()

    """

    def __init__(
        self,
        pupil,
        A=10.0,
        B=0.0,
        C=0.0,
        f=None,
        fknee=1.0,
        fmin=None,
        fmax=None,
        SR=0.0,
        units=u.m,
    ):

        # get the grid size
        Nx, Ny = pupil.shape

        # get the mask if it is a masked array
        if isinstance(pupil, np.ma.MaskedArray):
            mask = pupil.mask
        else:
            mask = np.zeros((Nx, Ny)).astype(bool)

        # compute the desired std for psd
        sfe_rms = self.sfe_rms(A, B, C, fknee, fmin, fmax)
        logger.debug(f"Desired std for PSD: {sfe_rms} {units}")

        # generate random field
        self.wfe = np.random.randn(Nx, Ny)

        # FFT
        ft_wfe = np.fft.fft2(self.wfe)

        dfx = f[0, 2] - f[0, 1]
        dfy = f[2, 0] - f[1, 0]

        # compute 2D PSD
        psd2d = A / (B + (f / fknee) ** C) / (2 * np.pi * f) * (dfx * dfy)

        # apply PSD to FT of random field
        ft_wfe *= np.sqrt(psd2d) * np.sqrt(Nx * Ny)

        # frequency mask
        idx = np.logical_or(f < fmin, f > fmax)
        ft_wfe[idx] = 0.0

        # inverse FFT
        self.wfe = np.fft.ifft2(ft_wfe).real

        # to masked array
        self.wfe = np.ma.masked_array(self.wfe, mask=mask)

        # current std
        current_std = self.wfe.std()
        logger.debug(f"Current std for PSD: {current_std} {units}")

        # add surface roughness
        self.wfe += SR * np.random.randn(Nx, Ny)

        # from sfe to wfe
        self.wfe *= 2

        # convert to meters using astropy.units
        self.wfe *= units.to(u.m)

    def __call__(self):

        return self.wfe

    @staticmethod
    def sfe_rms(A, B, C, f_knee, f_min, f_max):
        """
        Method to compute the rms of the surface error field (SFE) from the power spectral density (PSD).
        It uses sympy to evaluate the integral of the PSD.
        """
        # define symbols
        f = sympy.symbols("f")
        a, b, c, fknee, fmin, fmax = sympy.symbols("a b c fknee fmin fmax")

        # define PSD
        expr = a / (b + (f / fknee) ** c)

        # evaluate integral using sympy
        integral = sympy.integrate(expr, (f, fmin, fmax))

        # substitute values
        integral = integral.subs(
            {a: A, b: B, c: C, fknee: f_knee, fmin: f_min, fmax: f_max}
        )

        # desired std for psd
        sfe_rms = np.sqrt(float(integral))

        return sfe_rms
